/* Monitor configuration file for CSB250 board.
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author: Ed Sutter
 * email:  esutter@lucent.com
 * phone:  908-582-2351
 */

#ifndef CONFIG_H
#define CONFIG_H

#define CPU_BE

#define MONSTACKSIZE	(8*1024)

/* USE_TWO_TFS_DEVICES:
 * The memory map of CSB250 is a bit unusual.  This is because the reset
 * vector address (0xbfc00000) is right in the middle of the upper flash
 * bank; hence, the monitor binary is in the middle as well.
 *
 * TFS can be configured to support miltiple TFS "devices".  These TFS 
 * "devices" do not necessarily map to physical devices.
 *
 * If this macro is defined, then TFS will be configured with a //A
 * device and a //B device (see tfsdev.h).  The //A device covers the
 * space starting at flash bank 1 and ending below the base of the monitor
 * binary and //B being the space in flash bank 1 that is above the monitor
 * binary.
 * If this macro is not defined, then TFS will be configured with a single
 * //FLASH device that will occupy only the flash space above the monitor
 * binary.
 *
 * Note that for this CSB250 port, the flash in bank 0 is left for
 * configuration by the user.
 */
#define USE_TWO_TFS_DEVICES

/* FLASHRAM_BASE
 * If defined, then this points to a 1Mbyte block of ram to be allocated
 * to TFS for volatile storage.
 * Note that if this is not used, then the flashram.o object module should
 * be removed from the makefile's object list.
 */
#define FLASHRAM_BASE	0xA0200000

/* Specify CPU/PLATFORM type and name so that common code can be used
 * for a similar cpu, on different platforms.
 * The 'TYPE' definition is used for ifdefs in the code and the 'NAME'
 * is used for printfs in the code.
 */
#define CPU_NAME        "AMD Au1500"
#define PLATFORM_NAME   "Cogent CSB250"

/* Override the app ram base to set it to 3Mbyte.  This reserves space
 * for umon, the shared PCI window and RAM-based TFS.
 */
#define APPRAMBASE_OVERRIDE 0xA0300000

/* Base-address of boot flash must be hard-coded to 0xbfc00000
 * Note that this is right in the middle of the flash device's address
 * space.
 */
#define BOOTROMBASE_OVERRIDE 0xBFC00000

/* XBUFCNT & RBUFCNT:
 *  Number of transmit and receive buffers allocated to ethernet.
 *  The total of XBUFCNT+RBUFCNT should not exceed MAXEBDS
 */
#define XBUFCNT 	8
#define RBUFCNT 	8
#define XBUFSIZE	2048
#define RBUFSIZE	2048

/* DEFAULT_ETHERADD & DEFAULT_IPADD:
 *  Refer to notes in ethernet.c function EthernetStartup() for details
 *  regarding the use of these definitions.
 */
#define     DEFAULT_ETHERADD "00:30:23:00:00:01"
#define     DEFAULT_IPADD	 "192.168.0.214"

/* Define command list headers to be used by the "help"
 * command...
 */
#define MON_CMDLIST_HEADER "\nMicro-Monitor Command Set:"
#define APP_CMDLIST_HEADER "\nApplication-Installed Command Set:"

/* Defining DONT_CENTER_MONHEADER eliminates the automatic centering
 * of the monitor's startup banner...
 */
#define DONT_CENTER_MONHEADER

/* LOOPS_PER_SECOND:
 * Approximately the size of a loop that will cause a 1-second delay.
 * This can be guestimated or modified with the sleep -c command at the
 * monitor command line.
 */
#define LOOPS_PER_SECOND    40000

/* Flash bank configuration:
 */
#ifdef FLASHRAM_BASE
#define FLASHBANKS              3
#else
#define FLASHBANKS				2
#endif
#define FLASH_BANK0_WIDTH       4
#define FLASH_BANK0_SIZE  	    0x01000000
#define FLASH_BANK0_BASE_ADDR	0xbe000000
#define FLASH_BANK1_BASE_ADDR	0xbf000000
#define FLASH_LARGEST_SECTOR    0x20000
#define FLASH_TIMEOUT			1000000

#if FLASHRAM_BASE
#define FLASHRAM_END			0xa02fffff
#define FLASHRAM_SECTORSIZE		0x010000
#define FLASHRAM_SECTORCOUNT	16
#define FLASHRAM_BANKNUM		2
#endif

/* TFS definitions:
 *  TFSSTART:       Base address in FLASH at which TFS starts.
 *  TFSEND:         End address of TFS in FLASH.
 *  TFSSPARE:       Location of sector that is used as the spare sector
 *                  by TFS for defragmentation.
 *  TFSSPARESIZE:   Size of the spare sector used by TFS for defragmentation.
 *  TFSSECTORCOUNT: Number of eraseable sectors that TFS covers, not including
 *                  the TFSSPARE sector.
 */

/* The 0xBFC00000 address is right in the middle of the flash device, so
 * we configure TFS with two "devices".  One below and one above.
 */
#define TFSSPARESIZE_A			FLASH_LARGEST_SECTOR
#define TFSSECTORCOUNT_A  		95
#define TFSSTART_A				FLASH_BANK1_BASE_ADDR
#define TFSEND_A				0xbfbdffff
#define TFSSPARE_A				TFSEND_A+1
#define TFSSPARESIZE_B			FLASH_LARGEST_SECTOR
#define TFSSECTORCOUNT_B  		27
#define TFSSTART_B        		0xbfc80000
#define TFSEND_B          		0xbffdffff
#define TFSSPARE_B        		TFSEND_B+1

#define TFS_EBIN_ELF    		1
#define TFS_VERBOSE_STARTUP		1
#define TFS_AUTOBOOT_ABORTABLE	    


/* FLASH_PROTECT_SIZE is used by the flash code to determine how much of the
   boot flash is software-protected.  It is usually something to
   cover the size of the monitor executable. */
#define FLASH_PROTECT_RANGE "224-225"

/* SYMFILE is the name of a file that, if present, the monitor uses to 
 * process symbols on the command line.  A symbol is a whitespace delimited
 * string prefixed by a percent sign (%).
 */
#define SYMFILE "symtbl"

/* Specify the size of the memory block (in monitor space) that is to be
 * allocated to malloc in the monitor.  Note that this size can be dynamically
 * increased through the HEAPBASE and HEAPSTART shell variables.
 */
#define ALLOCSIZE 1024*64

#define NO_MONLIB_I2C_STUBS

/* Specify inclusion of subsystems within the monitor here.
 */
#define INCLUDE_MEMCMDS         1
#define INCLUDE_EDIT            1
#define INCLUDE_DISASSEMBLER    0
#define INCLUDE_UNZIP           1
#define INCLUDE_ETHERNET        1
#define INCLUDE_ICMP			1
#define INCLUDE_TFTP            1
#define INCLUDE_DHCPBOOT        1
#define INCLUDE_TFS             1
#define INCLUDE_TFSCLI          1
#define INCLUDE_TFSAPI          1
#define INCLUDE_TFSSYMTBL       1
#define INCLUDE_TFSSCRIPT       1
#define INCLUDE_XMODEM          1
#define INCLUDE_LINEEDIT        1
#define INCLUDE_EE              1
#define INCLUDE_FLASH           1
#define INCLUDE_MALLOC          1
#define INCLUDE_SHELLVARS       1
#define INCLUDE_STRACE          1
#define INCLUDE_CAST            1
#define INCLUDE_STRUCT          1
#define INCLUDE_REDIRECT        1
#define INCLUDE_QUICKMEMCPY     1
#define INCLUDE_PROFILER        1
#define INCLUDE_BBC             1
#define INCLUDE_STOREMAC        0
#define INCLUDE_MEMTRACE        1
#define INCLUDE_TRAPCMD         1
#define INCLUDE_SYSCALLCMD      1
#define INCLUDE_POLLABLECLOCK   0
#define INCLUDE_VERBOSEHELP		1
#define INCLUDE_HWTMR			0
#define INCLUDE_PORTCMD			0
#define INCLUDE_SYSLOG			1
#define INCLUDE_USRLVL			1

/* Unique to the CSB250...
 * These two configurable items allow the keyboard and display code to
 * be removed because if they aren't used, they waste a lot of time at
 * startup.
 */
#define INCLUDE_PS2KBD			0
#define INCLUDE_DISPLAY			0

/* Inclusion of this next file will make sure that all of the above
 * inclusions are legal; and adjust where necessary.
 */
#include "inc_check.h"

#endif  /* CONFIG_H */
