//==========================================================================
//
// lh79520_uart.h
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         02/16/2003
// Description:  This file contains register base addresses and offsets
//				 and access macros for the LH795200 on-chip UARTS
//

#include "bits.h"

// Register offsets
#define UART_DR				0x000	// Transmit (write) and Receive (read) Data Register
#define UART_RSR			0x004	// Receive Status Register - Read only
#define UART_ECR			0x004	// Error Clear Status Register - Write only
#define UART_FR				0x018	// Flag Register - Read only
#define UART_LPR			0x020	// Low Power Counter Register - UART0 SIR Mode only
#define UART_IBRD			0x024	// Buad Rate Integer Register
#define UART_FBRD			0x028	// Baud Rate Fractional Register
#define UART_LCR_H			0x02c	// Line Control Register, High Byte
#define UART_LCR_L			0x030	// Line Control Register, Low Byte
#define UART_IFLS			0x034	// FIFO Interrupt Level Register
#define UART_IMSC			0x038	// Interrupt Mask Set/Clear Register
#define UART_RIS			0x03c	// Raw Interrupt Status Register
#define UART_MIS			0x040	// Masked Interrupt Status Register
#define UART_ICR			0x044	// Interrupt Clear Register
#define UART_TCR			0x080	// SIR Test Counter Register - UART0 SIR Mode only

// Bit Defines
// UART_DR - Data Register
#define UART_DR_DATA		0x000000ff	// data read/write
#define UART_DR_FE			0x00000100	// Framing Error - Read Only
#define UART_DR_PE			0x00000200	// Parity Error - Read Only
#define UART_DR_BE			0x00000400	// Break Error - Read Only
#define UART_DR_OE			0x00000800	// Overrun Error - Read Only

// Bit defines for the following 2 registers are the same
// UART_RSR - Receive Status Register - Read only
// UART_ECR - Error Clear Status Register - Write only
#define UART_ERR_FE			0x00000001	// Framing Error
#define UART_ERR_PE			0x00000002	// Parity Error
#define UART_ERR_BE			0x00000004	// Break Error
#define UART_ERR_OE			0x00000008	// Overrun Error

// UART_FR - Flag Register - Read only
#define UART_FR_BUSY		0x00000008	// UART Busy
#define UART_FR_RXFE		0x00000010	// UART Receive FIFO Empty
#define UART_FR_TXFF		0x00000020	// UART Transmit FIFO Full
#define UART_FR_RXFF		0x00000040	// UART Receieve FIFO Full
#define UART_FR_TXFE		0x00000080	// UART Transmit FIFO Empty

// UART_LPR - Low Power Counter Register - UART0 SIR Mode only
#define UART_LPR			0x020	// Low Power Counter Register - UART0 SIR Mode only

// UART_LCR_H - Line Control Register, High Byte
#define UART_LCR_H_BRK		0x00000001	// Send Line Break
#define UART_LCR_H_PEN		0x00000002	// Enable Parity
#define UART_LCR_H_EPS		0x00000004	// 0 = Odd Parity, 1 = Even Parity
#define UART_LCR_H_STP2		0x00000008	// 0 = 1 Stop Bit, 1 = 2 Stop Bits
#define UART_LCR_H_FEN		0x00000010	// Enable FIFO's
#define UART_LCR_H_LEN6		0x00000020	// 0b01 = 6 bits
#define UART_LCR_H_LEN7		0x00000040	// 0b10 = 7 bits
#define UART_LCR_H_LEN8		0x00000060	// 0b11 = 8 bits, 0b00 = 5 bits
#define UART_LCR_H_SPS		0x00000080	// Force Stuck Parity

// UART_LCR_L - Line Control Register, Low Byte

// WARNING
// !!! LH79520 Users Manual shows EN, SIREN and SIRLP bits at 2-4, but the
// LH79520 example code shows them at bits 0-2.  Actual operation confirms them
// at bits 0-2 !!!
// #define UART_LCR_L_EN	0x00000004	// UART Enable - PER USERS MANUAL
// #define UART_LCR_L_SIREN	0x00000008	// IrDA Mode Enable - UART0 Only - PER USERS MANUAL
// #define UART_LCR_L_SIRLP	0x00000010	// IrDA Low Power Mode - UART0 Only - PER USERS MANUAL

#define UART_LCR_L_EN		0x00000001	// UART Enable -  - PER ACTUAL CODE AND USE !!!!
#define UART_LCR_L_SIREN	0x00000002	// IrDA Mode Enable - UART0 Only - PER ACTUAL CODE AND USE 
#define UART_LCR_L_SIRLP	0x00000004	// IrDA Low Power Mode - UART0 Only - PER ACTUAL CODE AND USE 
#define UART_LCR_L_LBE		0x00000080	// Loopback Enable
#define UART_LCR_L_TXE		0x00000100	// Transmit Enable
#define UART_LCR_L_RXE		0x00000200	// Receive Enable

// UART_IFLS -  FIFO Interrupt Level Register
#define UART_IFLS_TX2		0x00000000	// Interrupt when 2 bytes left in Transmit FIFO
#define UART_IFLS_TX4		0x00000001	// Interrupt when 4 bytes left in Transmit FIFO
#define UART_IFLS_TX8		0x00000002	// Interrupt when 8 bytes left in Transmit FIFO
#define UART_IFLS_TX12		0x00000003	// Interrupt when 12 bytes left in Transmit FIFO
#define UART_IFLS_TX14		0x00000004	// Interrupt when 14 bytes left in Transmit FIFO
#define UART_IFLS_RX2		0x00000000	// Interrupt when 2 bytes in Receive FIFO
#define UART_IFLS_RX4		0x00000008	// Interrupt when 4 bytes in Receive FIFO
#define UART_IFLS_RX8		0x00000010	// Interrupt when 8 bytes in Receive FIFO
#define UART_IFLS_RX12		0x00000018	// Interrupt when 12 bytes in Receive FIFO
#define UART_IFLS_RX14		0x00000020	// Interrupt when 14 bytes in Receive FIFO

// Bit Defines for the following 4 registers are the same
// UART_IMSC - Interrupt Mask Set/Clear Register
// UART_RIS - Raw Interrupt Status Register
// UART_MIS - Masked Interrupt Status Register
// UART_ICR - Interrupt Clear Register
#define UART_INT_RX			0x00000010	// Receive Interrupt       
#define UART_INT_TX			0x00000020	// Transmit Interrupt
#define UART_INT_RT			0x00000040	// Receive Timeout Interrupt
#define UART_INT_FE			0x00000080	// Framing Error Interrupt
#define UART_INT_PE			0x00000100	// Parity Error Interrupt
#define UART_INT_BE			0x00000200	// Break Error Interrupt
#define UART_INT_OE			0x00000400	// Overrun Error Interrupt
#define UART_INT_ALL		0x000006f0	// all interrupts combined
