/* main.c:
 * This is a superset of the ports.c file that comes with the
 * XILINX source of XAPP058 Reference C Code v5.01.
 * 
 * 
 * This application requires that an XSVF file be created.  To do this,
 * first generate the JEDEC file, and used IMPACT to convert that to SVF,
 * then use the svf2xsvf501 tool to convert from SVF to XSVF...
 *
 * Following is text yanked from the readme that came with the Xilinx
 * package...
 *
 *
 ************************************************************************
 ***** Create SVF Files to Program Your Device:
 ************************************************************************

  XC9500/XC9500XL/XC9500XV/CoolRunner/CoolRunner-II/XC18V00/XCF00/FPGA/
  System ACE MPM/SC:
    - Use iMPACT 5.2.03i or later (available from the WebPACK at
      http://www.xilinx.com/sxpresso/webpack.htm)
    - See iMPACT Help.
    - Create an SVF file that contains all the necessary operations to
      program your target devices.
        - Operations for XC9500/XL, CoolRunner/II, XC18V00, XCF00:
            Erase, Program, Verify
                (Operation->Program + check Erase before programming & Verify)
        - Operations for FPGA:
            Program
                (Operation->Program [no other options required])

 ************************************************************************
 ***** Translating SVF to XSVF:
 ************************************************************************

  SVF2XSVF Basic Syntax:

    svf2xsvf501 [options] -i input.svf -o output.xsvf -a output.txt

      -i input.svf   = any input file name for the input SVF file.
      -o output.xsvf = any output file name for output XSVF file.
      -a output.txt  = any output file name for output text version
                       of the XSVF file.  (For debugging/analysis.)

      [options] may be zero or more of the following:
      -d            = delete pre-existing output files.
      -w            = delete pre-existing output files.
      -r N          = max "retry" count = N times.  (XC9500/XL/XV only)
      -fpga         = Special translation for FPGA bitstreams.  (FPGA only)
      -extensions   = add special TAP command extensions (CoolRunner/II only)
      -xwait        = Use the inline XWAIT XSVF command
					  (CoolRunner/II/XC18V00/XCF00)
      -minxstate    = Minimize XSTATE XSVF commands
      -v2           = backward compatible XSVF with XSVF v2 player.
      -xprintsrc N  = print various levels of comments into XSVF (N=0-3)


  SVF2XSVF Options Requirements for Each Family:

    XC9500/XL/XV:                      // No options required

    CoolRunner/II: -r 0 -extensions -xwait

    XC18V00/XCF00: -r 0                // 0 = zero
        (See readme_xc18v00_xcf00s.txt for more XC18V00/XCF00S information!!!)

    All Spartan-II/Spartan-3/Virtex+ Platform FPGAs:
                  -fpga                // automatically sets -r 0

    Virtex-II/IIPro with encrypted bitstream or keys only
                  -fpga -extensions -xwait

    All XC4000+/Spartan/Spartan-XL FPGAs:
                  -fpga -rlen 1024     // automatically sets -r 0

    System ACE MPM/SC: -r 0 -xwait

  Command-line Translation Examples:

    XC9500/XL/XV:
    svf2xsvf -d -i xc9536.svf -o xc9536.xsvf -a xc9536.txt

    CoolRunner/II:
    svf2xsvf -d -r 0 -extensions -xwait -i xcr3064xl.svf
             -o xcr3064xl.xsvf -a xcr3064xl.txt

    XC18V00/XCF00:
    svf2xsvf -d -r 0 -i prom.svf -o prom.xsvf -a prom.txt
    (See readme_xc18v00_xcf00s.txt for more XC18V00/XCF00S information!!!)

    Spartan-II/Spartan-3/Virtex+ Platform FPGAs:
    svf2xsvf -d -fpga -i xcv300.svf -o xcv300.xsvf -a xcv300.txt

    XC4000+/Spartan/Spartan-XL FPGAs:
    svf2xsvf -d -fpga -rlen 1024 -i fpga.svf -o fpga.xsvf -a fpga.txt

    System ACE MPM/SC:
    svf2xsvf -d -r 0 -xwait -i mpm.svf -o mpm.xsvf -a mpm.txt
 *
 *
 */

#include "config.h"
#if INCLUDE_XSVFCMD
#include "genlib.h"
#include "stddefs.h"
#include "tfs.h"
#include "tfsprivate.h"
#include "cli.h"
#include "micro.h"
#include "ports.h"

/*
 ************************************************************************
 ***** Code specific to the CSB472:
 ************************************************************************
 */
#define LOOPS_PER_MICROSECOND 20
#define TICK_TOCK 	100

#define GPIO0_OR	0xEF600700	/* R/W GPIO Output */
#define GPIO0_TCR	0xEF600704	/* R/W Three-State Control (0=high-z) */
#define GPIO0_ODR	0xEF600718	/* R/W GPIO Open Drain */
#define GPIO0_IR	0xEF60071c	/* R GPIO Input */
#define GPIO_TCK	0x04000000	/* GPIO 5 */
#define GPIO_TMS	0x02000000	/* GPIO 6 */
#define GPIO_TDO	0x01000000	/* GPIO 7 */
#define GPIO_TDI	0x00800000	/* GPIO 8 */
#define GPIO_LED1	0x00040000	/* GPIO 13 */
#define GPIO_LED0	0x00020000	/* GPIO 14 */

/*
 * For the CSB472, the following connections are applicable:
 *	TDI: GPIO8		(output)
 *	TDO: GPIO7		(input)
 *	TMS: GPIO6		(output)
 *	TCK: GPIO5		(output)
 */
#define TDI_SET()	(*(vulong *)GPIO0_OR |= GPIO_TDI)
#define TMS_SET()	(*(vulong *)GPIO0_OR |= GPIO_TMS)
#define TCK_SET()	(*(vulong *)GPIO0_OR |= GPIO_TCK)
#define ALL_SET()	(*(vulong *)GPIO0_OR |= (GPIO_TCK|GPIO_TDI|GPIO_TMS))

#define TDI_CLR()	(*(vulong *)GPIO0_OR &= ~GPIO_TDI)
#define TMS_CLR()	(*(vulong *)GPIO0_OR &= ~GPIO_TMS)
#define TCK_CLR()	(*(vulong *)GPIO0_OR &= ~GPIO_TCK)
#define ALL_CLR()	(*(vulong *)GPIO0_OR &= ~(GPIO_TCK|GPIO_TDI|GPIO_TMS))

#define TDO_GET()	(*(vulong *)GPIO0_IR & GPIO_TDO)

#define LED0_ON()	(*(vulong *)GPIO0_TCR |= GPIO_LED0)
#define LED0_OFF()	(*(vulong *)GPIO0_TCR &= ~GPIO_LED0)
#define LED1_ON()	(*(vulong *)GPIO0_TCR |= GPIO_LED1)
#define LED1_OFF()	(*(vulong *)GPIO0_TCR &= ~GPIO_LED1)

void
pinConfigure(void)
{
	*(vulong *)GPIO0_TCR |= (GPIO_TDI | GPIO_TMS | GPIO_TCK);
	*(vulong *)GPIO0_TCR &= ~GPIO_TDO;

	ALL_CLR();
	TMS_SET();
}

void
resetTap(void)
{
	TCK_SET();
	TMS_SET();
	pulseClock();
	pulseClock();
	pulseClock();
	pulseClock();
	pulseClock();
}

//#define PORT_VERBOSE

/*
 ************************************************************************
 ***** End code specific to the CSB472
 ************************************************************************
 */
	
char *XsvfHelp[] = {
	"Reprogram the on-board CPLD from an xsvf file or memory",
	"[-aq] {filename | addr}",
	"Options:",
	" -a  indicates that argument is address",
	" -q  quiet mode (no ticker)",
	0,
};

char *xsvfBase;
int xsvfVerbose;

int
XsvfCmd(int argc, char *argv[])
{
	TFILE *tfp;
	int	opt, file, quiet;

	file = 1;
	quiet = 0;
	xsvfBase = 0;
	xsvfVerbose = 1;

    while ((opt=getopt(argc,argv,"aq")) != -1) {
	    switch(opt) {
	        case 'a':		/* arg is an address */
				file = 0;
				break;
	        case 'q':	
				xsvfVerbose = 0;
				break;
		}
	}

	if (argc != optind+1)
		return(CMD_PARAM_ERROR);

	if (file) {
		tfp = tfsstat(argv[optind]);
		if (!tfp) {
			printf("%s: not found\n",argv[optind]);
			return(CMD_FAILURE);
		}
		xsvfBase = (char *)TFS_BASE(tfp);
	}
	else {
		xsvfBase = (char *)strtoul(argv[optind],0,0);
	}

	/* Configure the GPIO pins that correspond to the JTAG-Control lines.
	 */
	pinConfigure();

	/* Program the device(s):
	 */
	printf("Transferring data at 0x%lx to CPLD...\n",(long)xsvfBase);
	xsvfExecute();

	/* Unconfigure the pins (put those GPIO pins back to tri-state or
	 * input mode)...
	pinUnConfigure();
	 */

	/* All done:
	 */
	return(CMD_SUCCESS);
}

/*************************************************************************
 *
 * All of the remaining functions are used by the XILINX source code
 * and provide the target-specific interface needed to allow the
 * package to program a device via JTAG.  To keep this code independent,
 * macros are used, as they are defined in the real target specific
 * code above.
 */

/* setPort():
 *
 * This function requires 6 external functions (part of target-specific)
 * code) to set and/or clear the three JTAG bits...
 *	setTMS(), clrTMS()
 *	setTDI(), clrTDI()
 *	setTCK(), clrTCK()
 */
void
setPort(short p,short val)
{
	static int tickcount;

	switch(p) {
		case TMS:
#ifdef PORT_VERBOSE
			printf("TMS%d\n",val);
#endif
			if (val)
				TMS_SET();
			else
				TMS_CLR();
			break;
		case TDI:
#ifdef PORT_VERBOSE
			printf("TDI%d\n",val);
#endif
			if (val)
				TDI_SET();
			else
				TDI_CLR();
			break;
		case TCK:
#ifdef PORT_VERBOSE
			printf("TCK%d\n",val);
#endif
			if (val)
				TCK_SET();
			else
				TCK_CLR();
			break;
		default:
			printf("setPort() 'p' error\n");
			break;
    }
	if (++tickcount > TICK_TOCK) {
		if (xsvfVerbose)
			ticktock();
		tickcount = 0;
	}
}

/* readTDOBit():
 * Return state of TDO bit (0 or 1).
 */
uchar
readTDOBit(void)
{
	uchar val;

	if (TDO_GET())
		val = 1;
	else
		val = 0;

#ifdef PORT_VERBOSE
	printf("TDO%d\n",val);
#endif
	return(val);
}

/* pluseClock():
 * toggle tck LH
 */
void
pulseClock(void)
{
    setPort(TCK,0);  /* set the TCK port to low  */
    setPort(TCK,1);  /* set the TCK port to high */
}

/* readByte():
 * Transfer the next byte of data to *data.
 */
void
readByte(uchar *data)
{
	*data = *xsvfBase++;
}

/* waitTime():
 * Wait "at least" the specified number of microseconds.
 * While waiting, the clock should be pulsed to satisfy
 * Virtex_II TCK cycles.
 */
void
waitTime(long microsec)
{
	volatile int	i;

	/* To keep this simple (at the expense of speed), we assume
	 * that the call to pulseClock() will take at least .01 microseconds...
	 */
	microsec *= LOOPS_PER_MICROSECOND;
    for ( i = 0; i < microsec; ++i );
}
#endif	/* INCLUDE_XSVFCMD */
