//==========================================================================
//
//      ads7846.c
//
// Author(s):   Michael Kelly - Cogent Computer Systems, Inc.
// Date:        03/06/03
// Description:	ADS7846 Interface routines for CSB536FS/CSB935FS 
//
//==========================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "stddefs.h"
#include "mc9328mxl.h"
#include "cpu_gpio.h"
#include "ads7846.h"
#include "cli.h"

//--------------------------------------------------------------------------
// function prototypes
//
int ads_init(void);
int ads_rd(uchar ads_ctl);

extern void udelay(int delay);

#define SPI_CTRL_BIT_CNT(_x_)	((_x_ & 0xf) << 0)	// bit count - 1
#define SPI_CTRL_POL_0			0					// see the MC9328MXL mamual
#define SPI_CTRL_POL_1			BIT4				// for a description of polarity
#define SPI_CTRL_PHASE_0		0					// and phase
#define SPI_CTRL_PHASE_1		BIT5				// 
#define SPI_CTRL_SS_ALWAYS		0					// SS stays low between transfers
#define SPI_CTRL_SS_PULSE		BIT6				// SS returns to inactive between transfers
#define SPI_CTRL_SS_LOW			0					// SS is low true
#define SPI_CTRL_SS_HIGH		BIT7				// SS is high true
#define SPI_CTRL_BUSY			BIT8				// read = transfer is in progress
#define SPI_CTRL_START			BIT8				// write = start transfer
#define SPI_CTRL_ENABLE			BIT9				// enable the SPI module
#define SPI_CTRL_MASTER			BIT10				// place us in Master mode
#define SPI_CTRL_RDY_OFF		(0x00 << 11)		// ignore RDY input
#define SPI_CTRL_RDY_EDGE		(0x01 << 11)		// wait for RDY input falling edge
#define SPI_CTRL_RDY_LEVEL		(0x02 << 11)		// wait for RDY input level low
#define SPI_CTRL_SCK4			(0x00 << 13)		// SCK divisor from PERCLK2 = 4	
#define SPI_CTRL_SCK8			(0x01 << 13)		// SCK divisor from PERCLK2 = 8	
#define SPI_CTRL_SCK16			(0x02 << 13)		// SCK divisor from PERCLK2 = 16	
#define SPI_CTRL_SCK32			(0x03 << 13)		// SCK divisor from PERCLK2 = 32	
#define SPI_CTRL_SCK64			(0x04 << 13)		// SCK divisor from PERCLK2 = 64	
#define SPI_CTRL_SCK128			(0x05 << 13)		// SCK divisor from PERCLK2 = 128
#define SPI_CTRL_SCK256			(0x06 << 13)		// SCK divisor from PERCLK2 = 256
#define SPI_CTRL_SCK512			(0x07 << 13)		// SCK divisor from PERCLK2 = 512

#define PEN_MOVE_FILTER			15

//--------------------------------------------------------------------------
// ads_init()
//
// This routine sets up the MC9328MXL SPI port in Microwire mode
// (8 bit control, 16 bit data, 24 bit total).  It also turns on
// the ADS7846 pen interrupt via a dummy read.  Note that we assume
// that PERCLK2 is 12Mhz (HCLK/4).  
//
int ads_init()
{

	SPI1_REG(SPI_CTRL) = SPI_CTRL_BIT_CNT(7)		// 8-bit data mode
					   | SPI_CTRL_POL_1				// POL_1 and  PHASE_1 = SCLK parks high 
					   | SPI_CTRL_PHASE_1			// and rises in the middle of the data
					   | SPI_CTRL_SS_PULSE			// SS goes inactive between transfers
					   | SPI_CTRL_MASTER			// Master mode
					   | SPI_CTRL_RDY_OFF			// ignore SPI_RDY
					   | SPI_CTRL_ENABLE			// turn the module on
					   | SPI_CTRL_SCK128;			// 12Mhz(PERCLK2)/128 = approx. 94Khz

	// enable the ADS7846 so it can generate a touch interrupt.
	// this consists of reading any channel, but setting the
	// power down mode in the control byte to 00b.  note we
	// don't bother with the returned data
	if (ads_rd(ADS7846E_S | ADS7846E_PD_LPWR) == -1)
	{
		printf("Error returned from ads_rd(0x%02x)!\n", (ADS7846E_S | ADS7846E_PD_LPWR));
		return -1;
	}
	return 0;
}

//--------------------------------------------------------------------------
// ads_rd(uchar ads_ctl)
//
// This function writes the 8-bit control value to the ADS7846, then reads
// the 16 bit result.
//
int ads_rd(uchar ads_ctl)
{
	int timeout = 5000;
	uchar temp0, temp1;

	// the MC9328MXL only handles up to 16-bits per transfer
	// so we send 8-bits at a time to get our total of 24
	// enable the slave select
	GPIO_clr(TSEN);
	SPI1_REG(SPI_TXD) = ads_ctl;
	SPI1_REG(SPI_CTRL) |= SPI_CTRL_START;


	// wait for not busy, then read the data
	while (timeout)
	{
		if ((SPI1_REG(SPI_CTRL) & SPI_CTRL_BUSY) == 0) break;
		udelay(10);	// 24bits at 128Khz takes about 250usec
		timeout --;
	}
	
	if (timeout == 0) {
		printf("ads_rd tmt\n");
		return -1;
	}
	else 
	{
		// toss the first byte
		temp0 = SPI1_REG(SPI_RXD);

		SPI1_REG(SPI_TXD) = 0;
		SPI1_REG(SPI_CTRL) |= SPI_CTRL_START;
		while ((SPI1_REG(SPI_CTRL) & SPI_CTRL_BUSY));
		temp0 = SPI1_REG(SPI_RXD);

		SPI1_REG(SPI_TXD) = 0;
		SPI1_REG(SPI_CTRL) |= SPI_CTRL_START;
		while ((SPI1_REG(SPI_CTRL) & SPI_CTRL_BUSY));
		temp1 = SPI1_REG(SPI_RXD);

		// disable the slave select
		GPIO_set(TSEN);

//		return (((temp0 << 8) | temp1) >> 4);	// lower 4-bit are padding
		return ((temp0 << 8) | temp1);
	}

	return 0;
}

//--------------------------------------------------------------------------
char *adsHelp[] = {
	"Screen touch test/demo",
	"",
	"Detect screen touches and display the x,y values via",
	"the ADS7846 and MC9328MXL SPI port.",
	0
};

int
ads(int argc, char *argv[])
{
	uchar cbuf[2], color;
	int pen, i, verbose, width;
	int sum_x, sum_y, last_x, last_y, average_x, average_y, x,y;

	// init the SPI and ADS7846
	if (ads_init() == -1)
	{
		printf("Error intializing ADS7846!\n");
		return (CMD_FAILURE);
	}
	lcd_init();
	lcd_on();

	printf("Scribble away!\n");
	printf("'0-f' changes color each time pen transitions to screen\n");
	printf("'i' refreshes the screen\n");
	printf("'w' decreases width of pen\n");
	printf("'W' increases width of pen\n");
	printf("'v' toggles verbosity\n");
	printf("Any other character terminates the demo\n");

	pen = 0;
	cbuf[1] = 0;
	width = 3;
	color = 0x0f;
	verbose = 0;
	while (1)
	{
		if (gotachar())
		{
			cbuf[0] = getchar();
			switch(cbuf[0]) {
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
				case 'a':
				case 'b':
				case 'c':
				case 'd':
				case 'e':
				case 'f':
					color = strtol((char *)cbuf,0,16);	
					break;
				case 'i':
					lcd_init();
					lcd_on();
					break;
				case 'W':
					if (width < 5)
						width++;
					break;
				case 'w':
					if (width > 1)
						width--;
					break;
				case 'v':
					verbose = verbose ? 0 : 1;
					break;
				default:
					goto done;
			}
		}
		if (GPIO_tst(PIRQ) == 0)
		{
			if (verbose)
				printf("Pen Down....\n");
			pen = 1;
			last_x = last_y = 0;
			while(GPIO_tst(PIRQ) == 0) // keep reading until touch goes away
			{
				sum_x = sum_y = 0;

				// display every 4 samples
				for (i = 0; i < 4 ; i++)
				{
					sum_x += ads_rd(ADS7846E_S | ADS7846E_PD_LPWR | ADS7846E_ADD_DFR_X);
					sum_y += ads_rd(ADS7846E_S | ADS7846E_PD_LPWR | ADS7846E_ADD_DFR_Y);
				}
				average_x = sum_x/4;
				average_y = sum_y/4;

				/* If incoming coordinate is ever more than 'PEN_MOVE_FILTER'
				 * pixels away from the previous position, then assume it
				 * to be noise and don't display it.
				 */
				if ((average_x < (last_x + PEN_MOVE_FILTER)) &&
				    (average_x > (last_x - PEN_MOVE_FILTER)) &&
				    (average_y < (last_y + PEN_MOVE_FILTER)) &&
				    (average_y > (last_y - PEN_MOVE_FILTER))) 
				{
					x = average_x*2;
					y = PIXELS_PER_COL - (average_y * 3);
					if (width >= 1) {
						fb_set_pixel(x,y,color);
					}
					if (width >= 2) {
						fb_set_pixel(x+1,y,color);
						fb_set_pixel(x-1,y,color);
						fb_set_pixel(x,y+1,color);
						fb_set_pixel(x,y-1,color);
					}
					if (width >= 3) {
						fb_set_pixel(x+2,y,color);
						fb_set_pixel(x-2,y,color);
						fb_set_pixel(x,y+2,color);
						fb_set_pixel(x,y-2,color);
					}
					if (width >= 4) {
						fb_set_pixel(x+3,y,color);
						fb_set_pixel(x-3,y,color);
						fb_set_pixel(x,y+3,color);
						fb_set_pixel(x,y-3,color);
					}
					if (width >= 5) {
						fb_set_pixel(x+4,y,color);
						fb_set_pixel(x-4,y,color);
						fb_set_pixel(x,y+4,color);
						fb_set_pixel(x,y-4,color);
					}
				}
				if (verbose) {
					if ((average_x != last_x) || (average_y != last_y))
						printf("X = %04d, Y = %04d\n", average_x,average_y);
				}

				last_x = average_x;
				last_y = average_y;

			} // while pen is down
		}
		if (pen)
		{
			if (verbose)
				printf("Pen Up....\n");
			pen = 0;
		}
	}

done:		
	printf("Demo terminated\n");
	return(CMD_SUCCESS);
}
