/*
 *  pxa250_i2c.h
 *
 *  Copyright (C) 2003 Technology Rescue.
 *	Author : David de la Cruz.
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *	History:
 *	15.04.2003 : CSB226 adaptation.
 */
#ifndef _I2C_PXA_H_
#define _I2C_PXA_H_

/* ----- Control register bits ---------------------------------------- */

#define ICR_START	0x1					/* start bit */
#define ICR_STOP	0x2					/* stop bit */
#define ICR_ACKNAK	0x4					/* send ACK(0) or NAK(1) */
#define ICR_TB		0x8 				/* transfer byte bit */
#define ICR_MA		0x10				/* master abort */
#define ICR_SCLE	0x20				/* master clock enable */
#define ICR_IUE		0x40				/* unit enable */
#define ICR_GCD		0x80				/* general call disable */
#define ICR_ITEIE	0x100				/* enable tx interrupts */
#define ICR_IRFIE	0x200				/* enable rx interrupts */
#define ICR_BEIE	0x400				/* enable bus error ints */
#define ICR_SSDIE	0x800				/* slave STOP detected int enable */
#define ICR_ALDIE	0x1000  			/* enable arbitration interrupt */
#define ICR_SADIE	0x2000				/* slave address detected int enable */
#define ICR_UR		0x4000				/* unit reset */

/* ----- Status register bits ----------------------------------------- */

#define ISR_RWM         0x1		/* read/write mode */
#define ISR_ACKNAK      0x2		/* ack/nak status */
#define ISR_UB          0x4		/* unit busy */
#define ISR_IBB         0x8		/* bus busy */	
#define ISR_SSD         0x10		/* slave stop detected */
#define ISR_ALD         0x20		/* arbitration loss detected */
#define ISR_ITE         0x40            /* tx buffer empty */
#define ISR_IRF         0x80            /* rx buffer full */
#define ISR_GCAD        0x100		/* general call address detected */
#define ISR_SAD         0x200		/* slave address detected */
#define ISR_BED         0x400           /* bus error no ACK/NAK */



#define DEF_TIMEOUT             3
#define BUS_ERROR               (-EREMOTEIO)
#define ACK_DELAY               0       /* time to delay before checking bus error */
#define MAX_MESSAGES            65536   /* maximum number of messages to send */

#define I2C_SLEEP_TIMEOUT       2       /* time to sleep for on i2c transactions */
#define I2C_RETRY               (-2000) /* an error has occurred retry transmit */
#define I2C_TRANSMIT		1
#define I2C_RECEIVE		0
#define I2C_PXA_SLAVE_ADDR      0x1    /* slave pxa unit address */
#define I2C_ICR_INIT            (ICR_BEIE | ICR_IRFIE | ICR_ITEIE | ICR_GCD | ICR_SCLE) /* ICR initialization value */

/* ICR initialize bit values 
*                       
*  15. FM       0 (100 Khz operation)
*  14. UR       0 (No unit reset)
*  13. SADIE    0 (Disables the unit from interrupting on slave addresses 
*                                       matching its slave address)
*  12. ALDIE    0 (Disables the unit from interrupt when it loses arbitration 
*                                       in master mode)
*  11. SSDIE    0 (Disables interrupts from a slave stop detected, in slave mode)  
*  10. BEIE     0 (Disable interrupts from detected bus errors, no ACK sent)
*  9.  IRFIE    1 (Enable interrupts from full buffer received)
*  8.  ITEIE    1 (Enables the I2C unit to interrupt when transmit buffer empty)
*  7.  GCD      1 (Disables i2c unit response to general call messages as a slave) 
*  6.  IUE      0 (Disable unit until we change settings)
*  5.  SCLE     1 (Enables the i2c clock output for master mode (drives SCL)   
*  4.  MA       0 (Only send stop with the ICR stop bit)
*  3.  TB       0 (We are not transmitting a byte initially)
*  2.  ACKNAK   0 (Send an ACK after the unit receives a byte)
*  1.  STOP     0 (Do not send a STOP)
*  0.  START    0 (Do not send a START)
*
*/

#define I2C_ISR_INIT            0x7FF  /* status register init */
/* I2C status register init values 
 *
 * 10. BED      1 (Clear bus error detected)
 * 9.  SAD      1 (Clear slave address detected)
 * 7.  IRF      1 (Clear IDBR Receive Full)
 * 6.  ITE      1 (Clear IDBR Transmit Empty)
 * 5.  ALD      1 (Clear Arbitration Loss Detected)
 * 4.  SSD      1 (Clear Slave Stop Detected)
 */

#endif

ulong i2c_wr_device(uchar dev_addr, uchar reg, uchar data);
ulong i2c_rd_device(uchar dev_addr, uchar reg, uchar *data);
ulong i2c_transmit_to_device(uchar addr_cmd, uchar *data, uchar length);
ulong i2c_receive_from_device(uchar addr_cmd, uchar *data, uchar length);
ulong i2c_wr_byte(uchar data);
uchar i2c_rd_byte(void);
void i2c_start(unsigned long icr_set, unsigned long icr_reset);
void i2c_stop(unsigned long icr_set, unsigned long icr_reset);
void i2c_midbyte(unsigned long icr_set, unsigned long icr_reset);
void i2c_show_reg( void);

/*
 * Set this to zero to remove all the debug statements via dead code elimination.
 */
//#define I2C_DEBUG       1

/* lowlevel functions:
 */
extern unsigned long i2c_wr_byte(unsigned char);
extern unsigned char i2c_rd_byte(void);
extern void i2c_start(unsigned long, unsigned long);
extern void i2c_repeat_start(unsigned char, unsigned char);
extern void i2c_stop(unsigned long, unsigned long);
extern void i2c_midbyte(unsigned long, unsigned long);
extern void i2c_abort(void);
extern int i2c_wait_bus_not_busy(void);
extern int i2c_wait_unit_not_busy(void);
extern void i2c_transfer(int,int,int);
extern int i2c_check_status(ulong);
extern void i2c_init(void);
extern void i2c_show_reg(void);
