    .file "rom_reset.s"

/*
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:  Ed Sutter
 * email:   esutter@lucent.com 
 * phone:   908-582-2351
 * 
 *
 * Modified by Luis Torrico for the CSB726 - PXA270 Single Board.
 *
 * rom_reset.s:
 *
  */

#include "config.h"
#include "warmstart.h"
#include "pxa270.h"
#include "pxa270_msc.h"
    /*
     * Have a separate stack for each processor mode.
     */

    /* define sizes for each mode's stack */
    .equ FiqStackSz, 4096
    .equ IrqStackSz, 4096
    .equ AbtStackSz, 4096
    .equ UndStackSz, 4096
    .equ SysStackSz, 4096

    /* declare the stacks */
    .extern MonStack
    .global FiqStack
    .global IrqStack
    .global AbtStack
    .global UndStack
    .global SysStack        
    
    /* allocate the stacks */
    .comm   FiqStack, FiqStackSz    /* for the FIQ mode */
    .comm   IrqStack, IrqStackSz    /* for the IRQ mode */
    .comm   AbtStack, AbtStackSz    /* for the Abort mode */
    .comm   UndStack, UndStackSz    /* for the Undef mode */
    .comm   SysStack, SysStackSz    /* for the System mode */
    /* User mode has the same stack as system mode. */

/*********************************************************************/
    
    .extern start

    .global reset
    .global coldstart
    .global lukewarmstart
    .global warmstart
    .global ipaddr
    .global etheraddr

    .text
        
    /*
     * Exception table at address 0
     */ 

reset:  
    b coldstart
    b undefined_instruction
    b software_interrupt
    b abort_prefetch
    b abort_data
    b not_assigned
    b interrupt_request
    b fast_interrupt_request

#include "moncomptr.S"
#include "etheraddr.S"
#include "alttfsdevtbl.S"

/*********************************************************************/

    /*
     * At the end of the reset sequence, MMU, Icache, Dcache,
     * and write buffer are all disabled.
     * Also IRQs and FIQs are disabled in the processor's CPSR
     * The operating mode is SVC (supervisory mode), and the
     * PC is vectored at 0x00000000. A branch in 0x00000000
     * brings us directly here.
	 *
	 */

coldstart:
	ldr r0, =0x2001					/* allow access to all coprocessors */
	mcr 	p15,0,r0,c15,c1,0
	nop
	nop
	nop

	ldr 	r0, =0x00000078
	mcr 	p15,0,r0,c1,c0,0		/* Disable MMU, caches, write buffer */
	nop
	nop
	nop

	ldr 	r0, =0x00000000
	mcr     p15,0,r0,c8,c7,0   		/* flush TLB's */
	mcr     p15,0,r0,c7,c7,0   		/* flush Caches */
	mcr		p15,0,r0,c7,c10,4		/* Flush Write Buffer */
	nop
	nop
	nop

	mvn    r0, #0                   /* grant manager access to all domains */
	mcr    p15,0,r0,c3,c0,0  

	/*
     * Initialize the PXA270 interrupt controller
     */

    /* Mask off all IRQs and FIQs */

    ldr r1, =(INT_BASE | INT_ICMR)
    ldr r2, =0x0	        	/* mask everything off */
    str r2, [r1]  			/* set mask (write to ICMR) */

    /*
     * Initialize PXA270 Clock Register
     */
    ldr r1, =CLK_BASE			/* base of clock registers */

    ldr r2, =0x02000290		   /* processor running at 520.0MHz @1.3v with turbo (L=16, 2N=5)*/
    str r2, [r1, #CLK_CCCR]
	ldr r3, [r1, #CLK_CCCR]

    ldr r0, =0x00000003       /* enter FCS and turn on Turbo mode System Bus frequency = half of run-mode frequency*/
//    ldr r0, =0x0000000B       /* enter FCS and turn on Turbo mode and System Bus frequency = run-mode frequency*/
    mcr p14,0,r0,c6,c0,0

    /*
     * Initialize PXA270 Memory Controller
     */
    ldr r1, =MSC_BASE			/* base of memory controller */

    ldr r2, =0x5ed84ee8			/* settings from WinCE5.0 BSP */
//    ldr r2, =0x2ef110b8			/* original settings by Cogent */
    str r2, [r1, #MSC_MSC0]		/* you must read the MSC0/1 registers before they */      
	ldr r3, [r1, #MSC_MSC0]		/* will update and work properly */

								 
    ldr r2, =0x5cfc5ffc			/* settings from WinCE5.0 BSP */
//    ldr r2, =0x53dc3ffc			/* original settings by Cogent */
    str r2, [r1, #MSC_MSC1]
    ldr r3, [r1, #MSC_MSC1]
    
    ldr r2, =0x5ed05fd4			/* settings from WinCE5.0 BSP */
//    ldr r2, =0x06690000			/* original settings by Cogent */
    str r2, [r1, #MSC_MSC2]
    ldr r3, [r1, #MSC_MSC2]

    ldr r2, =0x03ca4fff			/* build MDREFR in a specific order */
    str r2, [r1, #MSC_MDREFR]
    
    ldr r2, =0x03ca4030			/* REF Rate = (64MS/4096 Rows)/32 = 48 */
    str r2, [r1, #MSC_MDREFR]
    
    ldr r2, =0x03cd2030			/* Set K0RUN, K1RUN and K2RUN, set K0DB2, bit14 & bit 17 to 0 */
    str r2, [r1, #MSC_MDREFR]
    
    ldr r2, =0x038d2030			/* Clear Self Refresh */
    str r2, [r1, #MSC_MDREFR]
    
    ldr r2, =0x038db030			/* Set E0PIN and E1PIN */
    str r2, [r1, #MSC_MDREFR]
    
    ldr r2, =0x09cc19D0			/* Set CNFG, but don't enable just yet */
    str r2, [r1, #MSC_MDCNFG]

	/* pause for ~200usec for SDRAM to init */

	ldr		r3, =0x400			/* loop count */

do_delay:
	nop
	nop
	subs	r3,r3,#1
	bne		do_delay
	nop
	nop
	nop

	/* force 9 cycles to SDRAM */
	ldr     r2, =0xa0000000
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	str     r2, [r2]
	
	/* now enable SDRAM */

//    ldr r2, =0x89DB09DB			/* Set MDENX bit to 1 (large 1-Gbyte memory) for 128M SDRAM */
    ldr r2, =0x19Cf19D3			/* settings from WinCE5.0 BSP for 128M DRAM dual chip select chips*/
    str r2, [r1, #MSC_MDCNFG]	

	/* and do an MRS */
    ldr r2, =0x00220022
    str r2, [r1, #MSC_MDMRS]

	/* toggle the user LED */
    ldr r1, =GPIO_BASE			/* base of gpio registers */
    ldr r2, =0x00000020			/* make GPIO5 an output */
    str r2, [r1, #GPIO_GPDR0]

    ldr r2, =0x00000020			/* drive GPIO5 low */
    str r2, [r1, #GPIO_GPCR0]



/********************************************************************/

midstart:
    ldr r0, =INITIALIZE
    
    /* fall-through to 'lukewarmstart' */

/********************************************************************/
    
lukewarmstart:  
    /* Save the argument to r11 */
    mov r11, r0

    /*
     * *** DO NOT TOUCH R11 ***
     */ 

    /*
     * Set-up the stack-pointers for all operating modes
     */

    /* FIQ mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x11           /* set FIQ mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(FiqStack + FiqStackSz - 4)    /* initialize the stack ptr */
    /* IRQ mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x12           /* set IRQ mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(IrqStack + IrqStackSz - 4)    /* initialize the stack ptr */
    /* Abort mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x17           /* set Abort mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(AbtStack + AbtStackSz - 4)    /* initialize the stack ptr */
    /* Undef mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x1b           /* set Undef mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(UndStack + UndStackSz - 4)    /* initialize the stack ptr */
    /* System mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x1f           /* set System mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(SysStack + SysStackSz - 4)    /* initialize the stack ptr */
    /* 'warmstart' will take us back to SVC mode
       stack for SVC mode will also be setup in warmstart */

    mov r0, r11     /* get argument back from r11 */
    b   warmstart

    
warmstart:
    /* Save the argument to r11 */
    mov r11, r0

    /*
     * *** DO NOT TOUCH R11 ***
     */ 


    /* Change (back) to SVC mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x13           /* set System mode bits */
    msr CPSR_c, r0              /* move back to CPSR */

    /* Set stack pointer to just a bit beyond MonStack so that we
	 * can call stkinit()...
	 */
    ldr sp, =(MonStack + MONSTACKSIZE + 64)
	bl stkinit

    /* Reset the stack pointer for the SVC mode (our current mode) */
    ldr sp, =(MonStack + MONSTACKSIZE - 4)

    /*
     * Restore argument which was saved to r11 and jump to
     * the C function start().
     */ 
    
    mov r0, r11
jump_to_c:
    bl start

    /* the C code should never return */
    b reset

.align 4


