//==========================================================================
//
//      mc9328mx21_lcd.c
//
// Author(s):   Michael Kelly - Cogent Computer Systems, Inc.
// Date:        05-27-2002
// Description:	Init Code for MC9328mx21 LCD Controller
// NOTE 		Only 16-bit mode has been tested!
//
//==========================================================================

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "mcf532x.h"
#include "warmstart.h"
#include "vga_lookup.h"
#include "font8x16.h"
#include "fb_draw.h"

//--------------------------------------------------------------------------
// function prototypes and externs
//
void lcd_init(void);
void lcd_on(void);
void lcd_off(void);

extern void udelay(int);
extern int GPIO_clr(int);
extern int GPIO_set(int);
extern int GPIO_tst(int);
extern int GPIO_out(int);
extern int GPIO_in(int);

// ADDED FOR WRITING CHARACTERS TO THE DISPLAY
void lcd_putchar(char c);
void lcd_writechar(uchar c);
void lcd_clr_row(int char_row);
void lcd_clr_scr(void);
void lcd_switch_buffer(void);

// globals to keep track of foreground, background colors and x,y position
int lcd_color_depth;		// 4, 8 or 16
int lcd_fg_color;			// 0 to 15, used as lookup into VGA color table
int lcd_bg_color;			// 0 to 15, used as lookup into VGA color table
int lcd_col;				// current column, 0 to COLS_PER_SCREEN - 1
int lcd_row;				// current row, 0 to (ROWS_PER_SCREEN * 2) - 1
int lcd_tst_mode = 0;
ulong lcd_fb_offset;	   	// current offset into frame buffer for lcd_putchar

// globals to hold background and foreground colors
uchar bcolor=0;		// vga black
uchar color=15;		// vga white

//--------------------------------------------------------------------------
// lcd_init
//
// This function sets up the MCF5329 LCD controller
//
void
lcd_init(void)
{
	ushort temp16;
	int i;

	lcd_color_depth = 16;
	lcd_fg_color = vga_lookup[LCD_FG_DEF];
	lcd_bg_color = vga_lookup[LCD_BG_DEF];

	/* Pin assignments for ports LCDDATAH, LCDDATAM, LCDDATAL, LCDCTLL */
	MCF_GPIO_PDDR_LCDDATAH = 0;
	MCF_GPIO_PDDR_LCDDATAM = 0;
	MCF_GPIO_PDDR_LCDDATAL = 0;
	MCF_GPIO_PDDR_LCDCTLL = 0;
	//MCF_GPIO_PDDR_LCDCTLH = MCF_GPIO_PDDR_LCDCTLH_PDDR_LCDCTLH0;
	//MCF_GPIO_PDDR_LCDCTLL = MCF_GPIO_PDDR_LCDCTLL_PDDR_LCDCTLL6
	//					  | MCF_GPIO_PDDR_LCDCTLL_PDDR_LCDCTLL5
	//					  | MCF_GPIO_PDDR_LCDCTLL_PDDR_LCDCTLL4
	//					  | MCF_GPIO_PDDR_LCDCTLL_PDDR_LCDCTLL3;
	
	MCF_GPIO_PAR_LCDDATA = MCF_GPIO_PAR_LCDDATA_PAR_LD17(0x3)
						 | MCF_GPIO_PAR_LCDDATA_PAR_LD16(0x3)
						 | MCF_GPIO_PAR_LCDDATA_PAR_LD15_8(0x3)
						 | MCF_GPIO_PAR_LCDDATA_PAR_LD7_0(0x3);
	MCF_GPIO_PAR_LCDCTL = MCF_GPIO_PAR_LCDCTL_PAR_ACD_OE
						| MCF_GPIO_PAR_LCDCTL_PAR_FLM_VSYNC
						| MCF_GPIO_PAR_LCDCTL_PAR_LP_HSYNC
						| MCF_GPIO_PAR_LCDCTL_PAR_LSCLK;

	// Set the LCD screen start address.  This should be an are of 4MB in SDRAM.
    //MCF_LCDC_LSSAR = MCF_LCDC_LSSAR_SSA(LCD_BUF_ADD);
    MCF_LCDC_LSSAR = LCD_BUF_ADD;

	// Set the LCD size. We use a WQVGA 480x272 by default.
    MCF_LCDC_LSR = MCF_LCDC_LSR_XMAX(PIXELS_PER_ROW/16)
				 | MCF_LCDC_LSR_YMAX(PIXELS_PER_COL);

    // Set the Panel Control Register
    MCF_LCDC_LPCR = MCF_LCDC_LPCR_TFT
                  | MCF_LCDC_LPCR_COLOR
                  | MCF_LCDC_LPCR_PBSIZ_8
                  | MCF_LCDC_LPCR_BPIX_16bpp
                  | MCF_LCDC_LPCR_FLM
                  | MCF_LCDC_LPCR_LPPOL
				  | MCF_LCDC_LPCR_CLKPOL
				  | MCF_LCDC_LPCR_ENDSEL
				  | MCF_LCDC_LPCR_SCLKIDLE
				  | MCF_LCDC_LPCR_SCLKSEL
                  | MCF_LCDC_LPCR_PCD(LCD_PCD);

    //  LCDC_HCR - Horizontal Configuration Register
    MCF_LCDC_LHCR = MCF_LCDC_LHCR_H_WIDTH(LCD_H_WIDTH)    // Pulse Width in pixels
                  | MCF_LCDC_LHCR_H_WAIT_1(LCD_H_FRONT)   // Front Porch
                  | MCF_LCDC_LHCR_H_WAIT_2(LCD_H_BACK);   // Back Porch
             
    //  LCDC_VCR - Vertical Configuration Register
    MCF_LCDC_LVCR = MCF_LCDC_LVCR_V_WIDTH(LCD_V_WIDTH)    // Pulse Width in pixels
                  | MCF_LCDC_LVCR_V_WAIT_1(LCD_V_FRONT)   // Front Porch
                  | MCF_LCDC_LVCR_V_WAIT_2(LCD_V_BACK);   // Back Porch

	// Set the left/right panning offset register
	MCF_LCDC_LPOR = 0; 

	// Set the Virtual page width register
    MCF_LCDC_LVPWR = (PIXELS_PER_ROW/2);
	
	// Set the LCD screen start address.  This should be an are of 4MB in SDRAM. 
    MCF_LCDC_LSSAR = LCD_BUF_ADD;

	// LCDC DMA Control Register should be set for Dynamic Burst
    //MCF_LCDC_LDCR = 0x80080004;
    MCF_LCDC_LDCR = MCF_LCDC_LDCR_HM(0x4)
				  | MCF_LCDC_LDCR_TM(0x8);

	// fill the screen with the background color
	switch (lcd_color_depth){
		case 4:
			temp16 = lcd_bg_color | lcd_bg_color << 4 | lcd_bg_color << 8 | lcd_bg_color << 12;
			break;
		case 8:
			temp16 = lcd_bg_color | lcd_bg_color << 8;
			break;
		default:
			temp16 = lcd_bg_color;
			break;
	}
	for (i = 0; i < (PIXELS_PER_COL * PIXELS_PER_ROW) * 4; i += 2){	
		LCD_BUF(i) = temp16;
	}

	if (StateOfMonitor == INITIALIZE)
		printf("MCF5329 LCD Initialization Complete.\n");

	return;
}

//--------------------------------------------------------------------------
// lcd_on
//
// This function turns on the MC9328MX21 LCD Controller and the Backlight Inverter
//
void lcd_on()
{
	// Make sure it's off first
//	lcd_off();

	// Enable the clock to the LCDC
	// This is done in the Miscellaneous Control Register of the CCM
	MCF_CCM_MISCCR |= MCF_CCM_MISCCR_LCD_CHEN; // bit 8 controls the LCD clock

	// Drive LCD backlight enable signal high.
    //GPIO_set(LCD_BKL);
	MCF_GPIO_PPDSDR_LCDCTLL |= MCF_GPIO_PPDSDR_LCDCTLL_PPDSDR_LCDCTLL6;
}

//--------------------------------------------------------------------------
// lcd_off
//
// This function turns off the LCD Controller and the LCD backlight
//
void lcd_off()
{
	// turn off LCD backlight
    udelay(10000);
    //GPIO_clr(LCD_BKL);
	MCF_GPIO_PCLRR_LCDCTLL |= MCF_GPIO_PCLRR_LCDCTLL_PCLRR_LCDCTLL6;
	
	// Disable the clock to the LCDC
	// This is done in the Miscellaneous Control Register of the CCM
	MCF_GPIO_PPDSDR_LCDCTLL &= ~MCF_CCM_MISCCR_LCD_CHEN; // bit 8 controls the LCD clock


}

char *lcd_tstHelp[] = {
    "MCF5329 LCD controller test",
    "-[n,x,d[4,8,16]]",
	"The user may set color depth to run the test at.",
	"The frame buffer R/W test will test all of the frame ",
	"buffer regardless of depth.",
    "Options...",
    "  -n    run test without keycheck - CAUTION: RESET SYSTEM TO STOP!",
	"  -d4   run test, force a depth of 4-bits/pixel",
	"  -d8   run test, force a depth of 8-bits/pixel",
	"  -d16  run test, force a depth of 16-bits/pixel",
	"  -x    init only, do not run frame buffer tests",
	"",
	"  No options, default to current mode and depth.",
	0
};

int lcd_tst(int argc,char *argv[])
{
	volatile ushort wr16, rd16;
	int i, x, opt;
	int no_wait = 0;
	int init_only = 0;
	char c;

	lcd_off();

    while ((opt=getopt(argc,argv,"clnsxd:4,8,16")) != -1) {
        switch(opt) {
			case 'd':	// set the color depth
                switch(*optarg) {
            	    case '4':
	        	        lcd_color_depth = 4;
						printf("Forcing 4bpp Mode!\n");
        	            break;
    	            case '8':
            	        lcd_color_depth = 8;
						printf("Forcing 8bpp Mode!\n");
                    	break;
	                default:	// test with 16bpp
	                    lcd_color_depth = 16;
						printf("Forcing 16bpp Mode!\n");
						break;
                }
	            break;
	        case 'n':   // no waiting for keypress - fastest operation
				no_wait = 1;
				printf("No Keypress Mode, Must Reset System to Stop!\n");
				break;
	        case 'x':   // init only 
				no_wait = 1;
				printf("Initializing LCD, Skipping testsp!\n");
				init_only = 1;
				break;
			default:	// test with current mode
				break;
		}
	}

	// get the new parameters into the LCD controller
	lcd_tst_mode = 1;	// don't print to the LCD while we are testing it

	lcd_init();
	lcd_on();

	if (init_only) return 0;

	printf("Frame Buffer R/W...");
	// do an address=data read/write test on the frame buffer
	// PIXELS_PER_COL * PIXELS_PER_ROW is the highest pixel.  
	// Multiply by bits_per_pixel (sed_color_depth), then
	// divide by 8 to get the actual byte count.
	for (i = 0; i < LCD_FB_SIZE(lcd_color_depth) + LCD_ROW_SIZE(lcd_color_depth); i += 2){	
		LCD_BUF(i) = i & 0xffff;
		rd16 = LCD_BUF(i);
		if(rd16 != (i & 0xffff)){
			printf("Fail at 0x%08x, WR 0x%08x, RD 0x%04lx!\n",LCD_BUF_ADD + i, i, (ulong)rd16);
			return -1;
		}
	} 

	printf("OK!, Press key to continue.\n");

	c = getchar();

	printf("Frame Buffer Start: 0x%08x, End 0x%08x\n",LCD_BUF_ADD, 
			LCD_BUF_ADD + LCD_FB_SIZE(lcd_color_depth) + LCD_ROW_SIZE(lcd_color_depth));
	if (no_wait)
	{
		printf("Begin Full Screen Color Test.\n");
		while(1){
			// fill the frame buffer with incrementing color values
			for (x = 0; x < 16; x++){
				switch (lcd_color_depth){
					case 4:  wr16 = x | x << 4 | x << 8 | x << 12; break;
					case 8:  wr16 = x | x << 8; break;
					default: wr16 = vga_lookup[x]; break;	// 16-bits bypasses the lookup table
				}
				for (i = 0; i < LCD_FB_SIZE(lcd_color_depth); i += 2){	
					LCD_BUF(i) = wr16;
				}
			} // for x
		} // while
	} // no_wait
	else
	{
		printf("Begin Full Screen Color Test, Press any key to go to next color, \'x\' to end.\n");
		while(1){
			// fill the frame buffer with incrementing color values
			for (x = 0; x < 16; x++){
				switch (lcd_color_depth){
					case 4:  wr16 = x | x << 4 | x << 8 | x << 12; break;
					case 8:  wr16 = x | x << 8; break;
					default: wr16 = vga_lookup[x]; break;	// 16-bits bypasses the lookup table
				}
				for (i = 0; i < LCD_FB_SIZE(lcd_color_depth) + LCD_ROW_SIZE(lcd_color_depth); i += 2){	
					LCD_BUF(i) = wr16;
				}
				c = getchar();
				if (c == 'x') goto lcd_tst_next;
			} // for x
		} // while
	} // else no keycheck test

	lcd_tst_next:

	// write a one pixel border around the screen
	lcd_fg_color = 15;	// VGA Bright White
	fb_draw_line2(LEFT, TOP, RIGHT, TOP, lcd_fg_color);
	fb_draw_line2(LEFT, TOP, LEFT, BOTTOM, lcd_fg_color);
	fb_draw_line2(LEFT, BOTTOM, RIGHT, BOTTOM, lcd_fg_color);	// bottom left to bottom right
	fb_draw_line2(RIGHT, TOP, RIGHT, BOTTOM, lcd_fg_color);	// bottom right to top right
	// draw an x
	fb_draw_line2(LEFT, TOP, RIGHT, BOTTOM, lcd_fg_color);
	fb_draw_line2(LEFT, BOTTOM, RIGHT, TOP, lcd_fg_color);
	// draw 3 circles at the center of the screen, one inside the other
	lcd_fg_color = 12;	// VGA Bright Red
	fb_draw_circle2(CENTER_X, CENTER_Y, 100, lcd_fg_color);
	lcd_fg_color = 10;	// VGA Bright Green
	fb_draw_circle2(CENTER_X, CENTER_Y, 66, lcd_fg_color);
	lcd_fg_color = 9;	// VGA Bright Blue
	fb_draw_circle2(CENTER_X, CENTER_Y, 33, lcd_fg_color);

	return 0;
}

// fb_set_pixel sets a pixel to the specified color.
// This is target specific and is called from the generic
// fb_draw.c functions
void fb_set_pixel(ulong X, ulong Y, uchar color)
{
   	// Make sure the specified pixel is valid.
#if 0
   	if((X < 0) || (X >= PIXELS_PER_ROW) || (Y < 0) || (Y >= PIXELS_PER_COL))
   	{   
		printf("fb_set_pixel() bad X (%ld) or Y (%ld)!\n", X, Y);
		return;
   	}
#else
	if (X < 0)
		X = 0;
	else {
		if (X >= PIXELS_PER_ROW)
			X = PIXELS_PER_ROW-1;
	}
	if (Y < 0)
		Y = 0;
	else {
		if (Y >= PIXELS_PER_COL)
			Y = PIXELS_PER_COL-1;
	}
#endif

	LCD_BUF(LCD_GET_PIXEL_ADD(X, Y)) = vga_lookup[color];
}

//-------------------------------------------------------------------------
//
// THIS SECTION WAS ADDED 08-01-08 TO TEST IF WE COULD DISPLAY CHARACTERS
// ON THE DISPLAY WITH THE MX21.
//
//-------------------------------------------------------------------------

// lcd_scroll()
//
// Because we are most likely running out of FLASH and probably also with
// cache disabled, a brute force memcpy of the whole screen would be very
// slow, even with reduced color depths.  Instead, we define a frame buffer
// that is twice the size of our actual display.
//
// With the double buffering, we always write to the first buffer, even
// when the second buffer is active.  This allows us to scroll by adjusting
// the starting and ending addresses in the SAM9263 by one row.  When we
// reach the end of our virtual buffer, we reset the starting and ending
// addresses to the first buffer.
//
void lcd_scroll(void) {

	lcd_row++;

	// clear the new row(s)
	lcd_clr_row(lcd_row);
	if (lcd_row > (ROWS_PER_SCREEN - 1)){
		lcd_clr_row(lcd_row - ROWS_PER_SCREEN);
	}
	// when lcd_y_pos is greater than ROWS_PER_SCREEN we just adjust the
	// start and end addresses in the SAM9263.  If it is equal to 2 *
	// ROWS_PER_SCREEN, we reset the start and end addresses to LCD_MEM_BASE.
	if (lcd_row > (ROWS_PER_SCREEN - 1))
	{
		if (lcd_row > ((ROWS_PER_SCREEN * 2) - 1))
		{
			lcd_fb_offset = 0x00;
			lcd_row = ROWS_PER_SCREEN - 1;
		}
		else
		{
			// calculate the new offset address of the frame buffer in words
			lcd_fb_offset += (LCD_GET_ADD(1, 0, lcd_color_depth));
		}

#ifdef LCD_DBG
	lcd_tst_mode = 1;
	printf("LCD New FB Offset = 0x%08lx.\n ", lcd_fb_offset);
	lcd_tst_mode = 0;
#endif
//		LCD_FDR(0x4) = LCD_BUF_ADD + lcd_fb_offset;				// new beginning of Frame Buffer in Memory
		MCF_LCDC_LSSAR = LCD_BUF_ADD + lcd_fb_offset;				// new beginning of Frame Buffer in Memory

	} // if (lcd_row > (ROWS_PER_SCREEN - 1))
	return;
		

} // 

//--------------------------------------------------------------------------
// lcd_clr_row()
//
// This routine writes the background color to the font row specified
//
void lcd_clr_row(int char_row){

	ulong lcd_mem_add;
	int i;
	ushort wr16;

	// clear the desired row
	lcd_mem_add = LCD_GET_ADD(char_row, 0, lcd_color_depth);

#ifdef LCD_DBG
	lcd_tst_mode = 1;
	printf("LCD Clear Row %d, FB Add 0x%08lx.\n ", char_row, lcd_mem_add);
	lcd_tst_mode = 0;
#endif

	switch (lcd_color_depth)
	{
		case 4:
			wr16 = ((lcd_bg_color << 12) | (lcd_bg_color << 8) | (lcd_bg_color << 4) | (lcd_bg_color << 0));
			for (i = 0; i < ((PIXELS_PER_ROW * FONT_HEIGHT) / 2); i += 2){
				LCD_BUF(lcd_mem_add) = wr16;
				lcd_mem_add += 2;
			} // for font_row
			break;
		case 8:
			wr16 = ((lcd_bg_color << 8) | (lcd_bg_color << 0));
			for (i = 0; i < (PIXELS_PER_ROW * FONT_HEIGHT); i += 2){
				LCD_BUF(lcd_mem_add) = wr16;
				lcd_mem_add += 2;
			} // for font_row
			break;
		case 16:
			wr16 = (lcd_bg_color << 0);
			for (i = 0; i < ((PIXELS_PER_ROW * FONT_HEIGHT) * 2); i += 2){
				LCD_BUF(lcd_mem_add) = wr16;
				lcd_mem_add += 2;
			} // for font_row
			break;
	} // switch lcd_color_depth
} // lcd_clr_row


//--------------------------------------------------------------------------
// lcd_putchar()
//
// This routine parses the character and calls lcd_writechar if it is a
// printable character
//
void lcd_putchar(char c){

	if (lcd_tst_mode == 1) return;

	// First parse the character to see if it is printable or an acceptable control
    switch (c) {
	    case '\r':
	        lcd_col = 0;
	        return;
	    case '\n':
	        lcd_scroll();
	        return;
	    case '\b':
	        lcd_col--;
	        if (lcd_col < 0)
	        {
	            lcd_row--;
	            if (lcd_row < 0) lcd_row = 0;
	            lcd_col = COLS_PER_SCREEN - 1;
		    }
			c = 0;		// erase the character
			lcd_writechar(c);
	        break;
	    default:
	        if (((uchar)c < FIRST_CHAR) || ((uchar)c > LAST_CHAR)) return; // drop anything we can't print
			c -= FIRST_CHAR;	// get aligned to the first printable character
			lcd_writechar(c);
			// advance to next column
			lcd_col++;
		    if (lcd_col == COLS_PER_SCREEN)
		    {
				lcd_col = 0;
			    lcd_scroll();
			}
			break;
    } 

} // lcd_putchar()

//--------------------------------------------------------------------------
// lcd_writechar()
//
// This routine writes the character to the screen at the current cursor
// location.
//
void lcd_writechar(uchar c) {

	ulong lcd_mem_add;
	int font_row, font_col;
	uchar font_data8;
	ushort wr16;

	// Convert the current row,col and color depth values 
	// into an address
	lcd_mem_add = LCD_GET_ADD(lcd_row, lcd_col, lcd_color_depth);

#ifdef LCD_DBG
	lcd_tst_mode = 1;
	printf("LCD writechar at row %d, col %d, FB Add 0x%08lx.\n ", lcd_row, lcd_col, lcd_mem_add);
	lcd_tst_mode = 0;
#endif

	if (FONT_WIDTH == 8)
	{
		switch (lcd_color_depth){
			case 4:
				// Now render the font by painting one font row at a time
				for (font_row = 0; font_row < FONT_HEIGHT; font_row++){
					// get the font row of data
					font_data8 = fonttbl[(c * FONT_HEIGHT) + font_row];
//					font_data8 = fonttbl[(c * FONT_HEIGHT) + font_row];
					for (font_col = 0; font_col < 8; font_col += 4)
					{
						// get a words worth of pixels
						wr16 = (((font_data8 & 0x80) ? lcd_fg_color << 0 : lcd_bg_color << 0)
							  | ((font_data8 & 0x40) ? lcd_fg_color << 4 : lcd_bg_color << 4)
							  | ((font_data8 & 0x20) ? lcd_fg_color << 8 : lcd_bg_color << 8)
							  | ((font_data8 & 0x10) ? lcd_fg_color << 12 : lcd_bg_color << 12));
						font_data8 = font_data8 << 4;
//						wr16 = (((font_data8 & 0x80) ? lcd_fg_color << 12 : lcd_bg_color << 12)
//							  | ((font_data8 & 0x40) ? lcd_fg_color << 8 : lcd_bg_color << 8)
//							  | ((font_data8 & 0x20) ? lcd_fg_color << 4 : lcd_bg_color << 4)
//							  | ((font_data8 & 0x10) ? lcd_fg_color << 0 : lcd_bg_color << 0));
//						font_data8 = font_data8 << 4;
						LCD_BUF(lcd_mem_add) = wr16;
						// if we are in the 2nd frame buffer, write to the 1st frame buffer also
						if (lcd_row > (ROWS_PER_SCREEN - 1)){
							LCD_BUF(lcd_mem_add - LCD_FB_SIZE(lcd_color_depth)) = wr16;
						}
						lcd_mem_add += 2;
					} // for font_col
					// go to the next pixel row
					lcd_mem_add += (LCD_ROW_SIZE(lcd_color_depth) - ((FONT_WIDTH * lcd_color_depth) / 8));
				} // for font_row
				break;

			case 8:
				// Now render the font by painting one font row at a time
				for (font_row = 0; font_row < FONT_HEIGHT; font_row++){
					// get the font row of data
					font_data8 = fonttbl[(c * FONT_HEIGHT) + font_row];
					for (font_col = 0; font_col < 8; font_col += 2)
					{
						// get a words worth of pixels
						wr16 = (((font_data8 & 0x80) ? lcd_fg_color << 0 : lcd_bg_color << 0)
							  | ((font_data8 & 0x40) ? lcd_fg_color << 8 : lcd_bg_color << 8));
						font_data8 = font_data8 << 2;
//						wr16 = (((font_data8 & 0x80) ? lcd_fg_color << 8 : lcd_bg_color << 8)
//							  | ((font_data8 & 0x40) ? lcd_fg_color << 0 : lcd_bg_color << 0));
//						font_data8 = font_data8 << 2;
						LCD_BUF(lcd_mem_add) = wr16;
						// if we are in the 2nd frame buffer, write to the 1st frame buffer also
						if (lcd_row > (ROWS_PER_SCREEN - 1)){
							LCD_BUF(lcd_mem_add - LCD_FB_SIZE(lcd_color_depth)) = wr16;
						}
						lcd_mem_add += 2;
					} // for font_col
					// go to the next pixel row
					lcd_mem_add += (LCD_ROW_SIZE(lcd_color_depth) - ((FONT_WIDTH * lcd_color_depth) / 8));
				} // for font_row
				break;

			default:	// 16bpp is the default
				// Now render the font by painting one font row at a time
				for (font_row = 0; font_row < FONT_HEIGHT; font_row++){
					// get the font row of data
					font_data8 = fonttbl[(c * FONT_HEIGHT) + font_row];
					for (font_col = 0; font_col < 8; font_col++)
					{
						// get a words worth of pixels
						wr16 = ((font_data8 & 0x80) ? lcd_fg_color : lcd_bg_color);
						font_data8 = font_data8 << 1;
						LCD_BUF(lcd_mem_add) = wr16;
						// if we are in the 2nd frame buffer, write to the 1st frame buffer also
						if (lcd_row > (ROWS_PER_SCREEN - 1)){
							LCD_BUF(lcd_mem_add - LCD_FB_SIZE(lcd_color_depth)) = wr16;
						}
						lcd_mem_add += 2;
					} // for font_col
					// go to the next pixel row
					lcd_mem_add += (LCD_ROW_SIZE(lcd_color_depth) - ((FONT_WIDTH * lcd_color_depth) / 8));
				} // for font_row
				break;
		} // switch lcd_color depth
	} // FONT_WIDTH == 8
	else
	{
		return;
	}
} // lcd_writechar()	


void lcd_clr_scr(void)
{
	int	row;

	/* Clear screen by clearing each lines. */
	for(row = 0; row < (ROWS_PER_SCREEN * 2); row++)
	{
		lcd_clr_row(row);
	}

	/* Reset cursor position. */
	lcd_row = 0;
	lcd_col = 0;
	lcd_fb_offset = 0x00;

}

/* clear_memory():
 * Clear n dwords memory from start address sp.
 * Return sp.
 */
int *
clear_memory(register int *sp, register int n)
{
	register int *sp0 = sp;

	while (--n >= 0)
		*sp++ = 0x0;
	return (sp0);
}

