#include "config.h"
#include "stddefs.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "cli.h"
#include "warmstart.h"
#include "xgpio.h"
#include "xparameters.h"

#if INCLUDE_PCI
#include "xpci.h"

void pciInit(void);
#else
#define pciInit()
#endif

#include "xparameters.h"
#include "xcache_l.h"
#include "xuartns550_l.h"


#define UART_BASEADDR XPAR_OPB_UART16550_0_BASEADDR
#define UART_CLOCK    XPAR_XUARTNS550_CLOCK_HZ


/* devInit():
 * As a bare minimum, initialize the console UART here using the
 * incoming 'baud' value as the baud rate.
 */
void
devInit(int baud)
{
	XUartNs550_SetBaud(UART_BASEADDR, UART_CLOCK, baud);
	XUartNs550_mSetLineControlReg(UART_BASEADDR, XUN_LCR_8_DATA_BITS);
	pciInit();
}

/* ConsoleBaudSet():
 * Provide a means to change the baud rate of the running
 * console interface.  If the incoming value is not a valid
 * baud rate, then default to 9600.
 * In the early stages of a new port this can be left empty.
 * Return 0 if successful; else -1.
 */
int
ConsoleBaudSet(int baud)
{
	XUartNs550_SetBaud(UART_BASEADDR,UART_CLOCK,baud);
	return(0);
}

/* target_console_empty():
 * Target-specific portion of flush_console() in chario.c.
 * This function returns 1 if there are no characters waiting to
 * be put out on the UART; else return 0 indicating that the UART
 * is still busy outputting characters from its FIFO.
 * In the early stages of a new port this can simply return 1.
 */
int
target_console_empty(void)
{
	if (XUartNs550_mIsTransmitEmpty(UART_BASEADDR))
		return(1);
	return(0);
}

/* target_putchar():
 * When buffer has space available, load the incoming character
 * into the UART.
 */
int
target_putchar(char c)
{
    XUartNs550_SendByte(UART_BASEADDR, c);
    return((int)c);
}

/* target_gotachar():
 * Return 0 if no char is avaialable at UART rcv fifo; else 1.
 * Do NOT pull character out of fifo, just return status. 
 */
int
target_gotachar(void)
{
	if (XUartNs550_mIsReceiveData(UART_BASEADDR))
        return(1);
    return(0);
}

/* target_getchar():
 * Assume there is a character in the UART's input buffer and just
 * pull it out and return it.
 */
int 
target_getchar(void)
{
    char    c;

    c = XUartNs550_mReadReg(UART_BASEADDR,XUN_RBR_OFFSET);
    return((int)c);
}

/* intsoff():
 */
ulong
intsoff(void)
{
	return(0);
}

/* intsrestore():
 */
void
intsrestore(ulong msr)
{
}

/* cacheInitForTarget():
 */
void
cacheInitForTarget(void)
{
	/* Enable Caches */
	XCache_EnableICache(0x8c000001);
	icacheInvalidate = (int(*)(char *,int))XCache_InvalidateICache;
}

/* target_reset():
 */
void
target_reset(void)
{
	extern void reset(void);
	reset();
}

/* If any CPU IO wasn't initialized in reset.S, do it here...
 * This just provides a "C-level" IO init opportunity. 
 */
void
initCPUio(void)
{
	/* ADD_CODE_HERE */
}

int boot_base;

#if INCLUDE_PCI
XPci xpci;
extern int pciVerbose;

void
pciInit(void)
{
	XStatus xresult;

	if ((xresult = XPci_Initialize(&xpci,0,0,0)) != XST_SUCCESS) {
		printf("XPci_Initialize() failed %d\n",xresult);
		return;
	}

    XPci_ConfigOut32(&xpci, 0, 0, 0,
		XPCI_HDR_COMMAND,
		XPCI_STATCMD_ERR_MASK |
		XPCI_STATCMD_MEM_EN |
		XPCI_STATCMD_BUSM_EN |
		XPCI_STATCMD_PARITY |
		XPCI_STATCMD_SERR_EN);

	XPci_ConfigOut32(&xpci, 0, 0, 0, XPCI_HDR_LATENCY,0);
}

int
pciCtrl(int interface, int cmd, unsigned long arg1,unsigned long arg2)
{
	if (pciVerbose)
		printf("pciCtrl(cmd=%d,arg1=0x%lx,arg2=0x%lx)\n",cmd,arg1,arg2);
	return(0);
}

ulong
pciCfgRead(int interface,int bus,int dev,int func,int regno)
{
	Xuint32 cfgaddr, data;

	cfgaddr = XPci_ConfigPack((unsigned)bus,(unsigned)dev,(unsigned)func);

	if (pciVerbose)
		printf("pciCfgRead(dev=%d,func=%d,reg=%d) (cfgadr: 0x%lx)\n",
			dev,func,regno,(long)cfgaddr);

	data = XPci_ConfigIn(&xpci,cfgaddr,(Xuint8)regno*4);
	return((ulong)data);
}

int
pciCfgWrite(int interface,int bus,int dev,int func,int regno,ulong value)
{
	Xuint32 cfgaddr;

	cfgaddr = XPci_ConfigPack((unsigned)bus,(unsigned)dev,(unsigned)func);
	if (pciVerbose) {
		printf("pciCfgWrite(dev=%d,func=%d,reg=%d,val=0x%lx) (cfgadr: 0x%lx)\n",
			dev,func,regno,value,(long)cfgaddr);
	}

	XPci_ConfigOut(&xpci, cfgaddr,(Xuint8)regno*4,(Xuint32)value);
	return(0);
}

void
pciShow(void)
{
	printf("PCI RegBaseAddr: 0x%lx\n",xpci.RegBaseAddr);
}

#endif

#if INCLUDE_PIC
char *picHelp[] = {
	"Talk to on-board PIC",
	"{string-to-pic}",
	0
};

int
picCmd(int argc, char *argv[])
{
	int i;
	char c, *cp;
	static char resp[64];

	XUartNs550_SetBaud(XPAR_OPB_UART16550_1_BASEADDR,UART_CLOCK,115200);
	XUartNs550_mSetLineControlReg(XPAR_OPB_UART16550_1_BASEADDR,
			XUN_LCR_8_DATA_BITS);

	if (argc != 2) 
		return(CMD_PARAM_ERROR);
		
	cp = argv[1];
	printf("Sending <%s> to PIC...\n",cp);

	/* Send string and appended CR...
	 */
	while(*cp)
	    XUartNs550_SendByte(XPAR_OPB_UART16550_1_BASEADDR, *cp++);
	XUartNs550_SendByte(XPAR_OPB_UART16550_1_BASEADDR, '\r');

	cp = resp;
	for(i=0;i<64;i++) {
		/* Wait for incoming character:
		 */
		while(!(XUartNs550_mIsReceiveData(XPAR_OPB_UART16550_1_BASEADDR))) {
			if (target_gotachar())
				goto done;
		}
    	c = XUartNs550_mReadReg(XPAR_OPB_UART16550_1_BASEADDR,
				XUN_RBR_OFFSET);

		if (c == '\r')
			break;

		resp[i] = c;
	}

done:
	resp[i] = 0;
	printf("Rcvd: <%s>\n",resp);
	return(CMD_SUCCESS);
}
#endif


/* main():
 * For the ML403 build (as of Aug 2005), ld is not used (see makefile);
 * hence, main() is the first function called after crt0...
 * Here we call mySPInit() just to get the stack pointer inside MonStack[].
 * Then we jump into umon_start(), which is just start() (in start.c),
 * with a name change to avoid conflict with crt0's start() function.
 */
int
main()
{
	extern void umon_start(int);
	extern void mySPInit(void);
	
	mySPInit();
	start(INITIALIZE);
	puts("start returned!\n");
	return(0);
}
